/* -*- C++ -*- */
#include <windows.h>
#include <winbase.h>
#include <stdio.h>

#include "escape.h"
#include "color.h"
#include "code.h"
#include "charcode.h"

#define FOREGROUND_WHITE (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE)
#define EITHER_CTRL_PRESSED (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED)
#define EITHER_ALT_PRESSED (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED)

#define ESCAPE_LOG escape_log_file
FILE *ESCAPE_LOG;

  /****************************************/
 /*	         ֐`                */
/****************************************/

/****** ֐ ******/
void LookupEscapeSequence(void);
void InterpretEscapeSequence(int);

/****** ESCV[PX֐ ******/
typedef void (*esc_func)(void);
void EscapeFunctionAttribute1(void);
void EscapeFunctionAttribute2(void);
void EscapeFunctionAttribute3(void);
void EscapeFunctionLocate1(void);
void EscapeFunctionLocate2(void);
void EscapeFunctionInsertCharacter(void);
void EscapeFunctionDeleteCharacter(void);
void EscapeFunctionClearLine(void);
void EscapeFunctionClearRegion(void);
void EscapeFunctionUp(void);
void EscapeFunctionDown(void);
void EscapeFunctionRight(void);
void EscapeFunctionLeft(void);
void EscapeFunctionMove(int,int);
void EscapeFunctionInsertLine(void);
void EscapeFunctionDeleteLine(void);
void EscapeFunctionScrollUp(void);
void EscapeFunctionScrollDown(void);
void EscapeFunctionResetScrollArea(void);
void EscapeFunctionSetScrollArea(void);
void EscapeFunctionJISX0208(void);
void EscapeFunctionASCII(void);
void EscapeFunctionHanKata(void);
void EscapeFunctionDECPrivateModeReset(void);
void EscapeFunctionDECPrivateModeSet(void);
void EscapeFunctionDeviceStatusReport(void);
void EscapeFunctionEnterMode(void);
void EscapeFunctionEndMode(void);
void EscapeFunctionSaveCursorPosition(void);
void EscapeFunctionRestoreCursorPosition(void);
void EscapeFunctionSetTextParameters(void);
void EscapeFunctionDummy(void);

/****** 䕶֐ ******/
void FunctionLineFeed(void);
void FunctionBackSpace(void);
void FunctionHorizontalTab(void);

  /****************************************/
 /*	    X^eBbNϐ`         */
/****************************************/

//static HANDLE hstdout;

static SMALL_RECT scroll_rect={-1,-1,-1,-1};
static int current_line=-1;

#define ESC_TYPES 40
static const int ESC_NUMARG_TABLE[ESC_TYPES]={
    1,2,3,1, 2,1,1,1, 1,1,1,1, 1,1,1,1, 
    1,1,2,3, 4,1,1,0, 2,0,0,0, 0,0,0,0,
    0,0,0,1, 0,0,1,2,
};

//-O LĂ̂ JISX211 ɒ`ẮB
static const char * ESC_TABLE[ESC_TYPES]={
    "[\001m",                 //SGR-SetGraphicRendition Agr[g
    "[\001;\001m",            //SGR-SetGraphicRendition Agr[g
    "[\001;\001;\001m",       //SGR-SetGraphicRendition Agr[g
    "[\001H",                 //CUP-CursorPosition J[\ʒuݒ
    "[\001;\001H",            //CUP-CursorPosition J[\ʒuݒ
    "[\001@",		      //ICH-InsertCharacter }
    "[\001P",		      //DCH-DeleteCharacter 폜
    "[\001K",                 //EL-EraseInLine s
    "[\001J",                 //ED-EraseInPage ̈
    "[\001A",                 //CUU-CursorUp    J[\
    "[\001B",                 //CUD-CursorDown  J[\
    "[\001C",                 //CUF-CursorRight J[\E
    "[\001D",                 //CUB-CursorLeft  J[\
    "[\001L",		      //IL-InsertLine s}
    "[\001M",		      //DL-DeleteLine s폜
    "[\001S",                 //SU-ScrollUp XN[H 
    "[?\001h",                // DEC Private Mode Set
                              // [?1h ́AJ[\[1B]O? o
    "[?\001l",                // DEC Private Mode Reset
    "[?\001;\001l",	      // [?1l ́AJ[\[1B][A`[1B][Do
    "[?\001;\001;\001l",      // 
    "[?\001;\001;\001;\001l", // 
    "[\001h",		      // 4ȂEnter Insert Mode(im)
    "[\001l",		      // 4ȂEnd Insert Mode(ei)
    "[r",                     // Reset Scroll Area
    "[\001;\001r",            // Set Scroll Area
    "$B",                     // JIS X 0208-1983 character set
    "$@",		      // JIS X 0208-1983(ّ̎) character set
    "(B",                     // ASCII (ANSI X3.4-1986) character set
    "(J",		      // JIS X 0201 e character set
    "(I",                     // pJ^JiH
    "=",                      // eL[[1B]O?o
    ">",                      // eL[0`9o
    "D",                      // Scroll Up  (forward) ISO͖`
    "M",                      // Scroll Down(reverse) ISOƈႤ
    "H",                      // HTS-CharacterTabulationSet 
    "[\001g",                 // Tab Clear 
    "7",		      // Save Cursor Position 
    "8",		      // Restore Cursor Position 
    "[\001n",		      // DSR-DeviceStatusReport uԕ(v)
    "]\001;\002\007",	      // Set Text Parameters
};



static const esc_func ESC_FUNC_TABLE[ESC_TYPES]={
    EscapeFunctionAttribute2,
    EscapeFunctionAttribute2,
    EscapeFunctionAttribute2,
    EscapeFunctionLocate1,
    EscapeFunctionLocate2,
    EscapeFunctionInsertCharacter,
    EscapeFunctionDeleteCharacter,
    EscapeFunctionClearLine,
    EscapeFunctionClearRegion,
    EscapeFunctionUp,
    EscapeFunctionDown,
    EscapeFunctionRight,
    EscapeFunctionLeft,
    EscapeFunctionInsertLine,
    EscapeFunctionDeleteLine,
    EscapeFunctionDummy,
    EscapeFunctionDECPrivateModeSet,
    EscapeFunctionDECPrivateModeReset,
    EscapeFunctionDECPrivateModeReset,
    EscapeFunctionDECPrivateModeReset,
    EscapeFunctionDECPrivateModeReset,
    EscapeFunctionEnterMode,
    EscapeFunctionEndMode,
    EscapeFunctionResetScrollArea,
    EscapeFunctionSetScrollArea,
    EscapeFunctionJISX0208,
    EscapeFunctionJISX0208,
    EscapeFunctionASCII,
    EscapeFunctionASCII,
    EscapeFunctionHanKata,
    EscapeFunctionDummy,
    EscapeFunctionDummy,
    EscapeFunctionScrollUp,
    EscapeFunctionScrollDown,
    EscapeFunctionDummy,
    EscapeFunctionDummy,
    EscapeFunctionSaveCursorPosition,
    EscapeFunctionRestoreCursorPosition,
    EscapeFunctionDeviceStatusReport,
    EscapeFunctionSetTextParameters,
};

#define OUTPUT_BUF_SIZE 256
static int FlashCursor = 0;
static char OUTPUT_BUF[OUTPUT_BUF_SIZE];
#ifdef DEBUG
static char DebugDummy;
#endif
static int  OUTPUT_BYTES=0;
static COORD CursorPosition;
#define ESC_BUF_SIZE 256
#define ESC_TEXT_ARG_SIZE 256
static char ESC_FLAG=0;
static char ESC_BUF[ESC_BUF_SIZE];
static char ESC_TEXT_ARG[ESC_TEXT_ARG_SIZE];
static int ESC_ARG[4];
static int ESC_ARGC;
static int ESC_BYTES;
static int ESC_CODING=0;
static int ESC_KANJI1=0;
static int ESC_SJIS1=0;
static int ESC_CURSOR=0;
static char ESC_INSERT_MODE=0;

  /**************************************/
 /*	    eXgpmain()           */
/**************************************/
/*
void main(void){
    hstdout=GetStdHandle(STD_OUTPUT_HANDLE);
    escprint("test\x1b[31mtest\x1b[m");
//    escprint("\x1b[H");
//    escprint("\x1b[;H");
//    escprint("\x1b[24;1H");
    escprint("\x1b[@");
    escprint("This is A Test!\n");
}
*/

  /**************************************/
 /*	 EscapeFunctions       */
/**************************************/

void EscapeFunctionAttribute1(void){}
void EscapeFunctionAttribute2(void){
    int arg,arg_index;
    int color,intensity;
    color=FG_WHITE;
    intensity=0;
    for(arg_index=0;arg_index<3;arg_index++){
	arg=ESC_ARG[arg_index];
	if(arg==0){
	    color=FG_WHITE;
	}else if(arg==1){
	    color|=FG_I;
	    intensity=1;
	}else if(arg==7){
	    //reverse
	    color<<=4;
	}else if(arg>=30){
	    arg-=30;
	    color=intensity?FG_I:0;
	    if(arg & 1){
		color|=FG_RED;
	    }
	    if(arg & 2){
		color|=FG_GREEN;
	    }
	    if(arg & 4){
		color|=FG_BLUE;
	    }
	}
    }
    SetConsoleTextAttribute(hstdout,(WORD)color);

}

void EscapeFunctionAttribute3(void){}
void EscapeFunctionLocate1(void){
    COORD coord;
    if(ESC_ARG[0]==-1){
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	coord.X=0;
	coord.Y=buffer_info.srWindow.Top;
	SetConsoleCursorPosition(hstdout,coord);
//	current_line=coord.Y;
    }else{
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	coord.X=ESC_ARG[0]-1;
	coord.Y=buffer_info.dwCursorPosition.Y;
	SetConsoleCursorPosition(hstdout,coord);
//	current_line=coord.Y;
    }
    current_line=-1;
}
void EscapeFunctionLocate2(void){
    COORD coord;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    if(ESC_ARG[0]==-1 && ESC_ARG[1]==-1){
	coord.X=0;
	coord.Y=buffer_info.srWindow.Top;
    }else{
	if(ESC_ARG[1]==-1){
	    coord.X=buffer_info.dwCursorPosition.X;
	}else{
	    coord.X=ESC_ARG[1]-1;
	    if(coord.X>buffer_info.srWindow.Right){
		coord.X=buffer_info.srWindow.Right;
	    }
	}
	if(ESC_ARG[0]==-1){
	    coord.Y=buffer_info.dwCursorPosition.Y;
	}else{
	    coord.Y=buffer_info.srWindow.Top+ESC_ARG[0]-1;
	    if(coord.Y>=buffer_info.srWindow.Bottom){
		coord.Y=buffer_info.srWindow.Bottom-1;
	    }
	}
    }
//    current_line=coord.Y;
    current_line=-1;
    SetConsoleCursorPosition(hstdout,coord);
}

void EscapeFunctionInsertCharacter(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect;
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    move_rect.Top=buffer_info.dwCursorPosition.Y;
    move_rect.Bottom=move_rect.Top;
    move_rect.Left=buffer_info.dwCursorPosition.X;
    move_rect.Right=buffer_info.dwSize.X-1;
    destination.X=move_rect.Left+ESC_ARG[0];
    destination.Y=move_rect.Top;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
}

void EscapeFunctionDeleteCharacter(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect;
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    move_rect.Top=buffer_info.dwCursorPosition.Y;
    move_rect.Bottom=move_rect.Top;
    move_rect.Left=buffer_info.dwCursorPosition.X;
    move_rect.Right=buffer_info.dwSize.X-1;
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    destination.X=move_rect.Left-ESC_ARG[0];
    destination.Y=move_rect.Top;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
}

void EscapeFunctionClearLine(void){
    COORD coord;
    DWORD numwritten;
    int arg;
    arg=ESC_ARG[0];
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    if(arg==-1 || arg==0){
	coord=buffer_info.dwCursorPosition;
	FillConsoleOutputCharacter(hstdout,' ',buffer_info.dwSize.X-coord.X,
				   coord,&numwritten);
	FillConsoleOutputAttribute(hstdout,FG_WHITE,buffer_info.dwSize.X-coord.X,
				   coord,&numwritten);
	SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
    }else if(arg==1){
	coord.X=0;
	coord.Y=buffer_info.dwCursorPosition.Y;
	FillConsoleOutputCharacter(hstdout,' ',buffer_info.dwCursorPosition.X,
				   coord,&numwritten);
	FillConsoleOutputAttribute(hstdout,FG_WHITE,buffer_info.dwCursorPosition.X,
				   coord,&numwritten);
	SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
    }else if(arg==2){
	coord.X=0;
	coord.Y=buffer_info.dwCursorPosition.Y;
	FillConsoleOutputCharacter(hstdout,' ',buffer_info.dwSize.X,
				   coord,&numwritten);
	FillConsoleOutputAttribute(hstdout,FG_WHITE,buffer_info.dwSize.X,
				   coord,&numwritten);
	SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
    }
}
void EscapeFunctionClearRegion(void){
    COORD coord;
    DWORD numwritten;
    int height;
    int i;
    if(ESC_ARG[0]==0 || ESC_ARG[0]==-1){ // J[\ʒuȍ~
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	coord.X=0;
	coord.Y=buffer_info.dwCursorPosition.Y;
	height=buffer_info.srWindow.Bottom-coord.Y+1;
	FillConsoleOutputCharacter(hstdout,' ',buffer_info.dwSize.X*height,
				   coord,&numwritten);
	FillConsoleOutputAttribute(hstdout,FG_WHITE,buffer_info.dwSize.X*height,
				   coord,&numwritten);
//	current_line=buffer_info.dwCursorPosition.Y;
	SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
    }else if(ESC_ARG[0]==1){	// J[\ʒuȑO
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	coord.X=0;
	coord.Y=buffer_info.srWindow.Top;
	height=buffer_info.dwCursorPosition.Y-buffer_info.srWindow.Top+1;
	FillConsoleOutputCharacter(hstdout,' ',buffer_info.dwSize.X*height,
				   coord,&numwritten);
	FillConsoleOutputAttribute(hstdout,FG_WHITE,buffer_info.dwSize.X*height,
				   coord,&numwritten);
//	current_line=buffer_info.dwCursorPosition.Y;
	SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
    }else if(ESC_ARG[0]==2){	// Sʏ
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	coord.X=0;
	coord.Y=buffer_info.srWindow.Top;
	SetConsoleCursorPosition(hstdout,coord);

	height=buffer_info.srWindow.Bottom-buffer_info.srWindow.Top+1;
	FillConsoleOutputCharacter(hstdout,' ',buffer_info.dwSize.X*height,
				   coord,&numwritten);
	FillConsoleOutputAttribute(hstdout,FG_WHITE,buffer_info.dwSize.X*height,
				   coord,&numwritten);
//	current_line=buffer_info.dwCursorPosition.Y;
	SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
    }

}

void EscapeFunctionUp(void){
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    EscapeFunctionMove(0,-ESC_ARG[0]);
}
void EscapeFunctionDown(void){
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    EscapeFunctionMove(0,ESC_ARG[0]);
}
void EscapeFunctionRight(void){
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    EscapeFunctionMove(ESC_ARG[0],0);
}
void EscapeFunctionLeft(void){
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    EscapeFunctionMove(-ESC_ARG[0],0);
}

void EscapeFunctionMove(int x,int y){
    COORD coord;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    coord=buffer_info.dwCursorPosition;
    if(current_line!=-1){
	coord.X=buffer_info.dwSize.X-1;
	coord.Y=current_line;
    }
    coord.X+=x;
    if(coord.X<0)coord.X=0;
    else if(coord.X>=buffer_info.dwSize.X)coord.X=buffer_info.dwSize.X-1;
    coord.Y+=y;
    if(coord.Y<0)coord.Y=0;
    else if(coord.Y>=buffer_info.dwSize.Y)coord.Y=buffer_info.dwSize.Y-1;
//    current_line=coord.Y;
    current_line=-1;
    SetConsoleCursorPosition(hstdout,coord);
}

void EscapeFunctionInsertLine(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect;
    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    move_rect.Top=buffer_info.dwCursorPosition.Y;
    if(scroll_rect.Top==-1){
	move_rect.Bottom=buffer_info.srWindow.Bottom;
    }else{
	move_rect.Bottom=buffer_info.srWindow.Top+scroll_rect.Bottom;
    }
    move_rect.Left=0;
    move_rect.Right=buffer_info.dwSize.X-1;
    destination.X=0;
    destination.Y=move_rect.Top+ESC_ARG[0];
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
    SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
}

#if 0
void EscapeFunctionDeleteLine(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect,source_rect;
    COORD move_rect_size;
    SHORT source_rect_height;

    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    if(scroll_rect.Top==-1){
	move_rect.Top=buffer_info.dwCursorPosition.Y;
	move_rect.Bottom=buffer_info.srWindow.Bottom;
    }else{
	move_rect.Top=buffer_info.dwCursorPosition.Y;
	move_rect.Bottom=buffer_info.srWindow.Top+scroll_rect.Bottom;
    }
    move_rect.Left=0;
    move_rect.Right=buffer_info.dwSize.X-1;
    destination.X=0;
    destination.Y=move_rect.Top;
    source_rect.Top=move_rect.Top+ESC_ARG[0];
    source_rect.Bottom=move_rect.Bottom;
    source_rect.Left=move_rect.Left;
    source_rect.Right=move_rect.Right;
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,&source_rect,
			       &move_rect,destination,&char_info);
    move_rect_size.Y=move_rect.Bottom-move_rect.Top+1;
    if(move_rect_size.Y< ESC_ARG[0] * 2){
      move_rect_size.X=move_rect.Right-move_rect.Left+1;
      source_rect_height=source_rect.Bottom-source_rect.Top+1;
      destination.Y+=source_rect_height;
      FillConsoleOutputCharacter(hstdout,' ',
	  (move_rect_size.Y - source_rect_height * 2)*move_rect_size.X,
	  destination,NULL);
      FillConsoleOutputAttribute(hstdout,0,
	  (move_rect_size.Y - source_rect_height * 2)*move_rect_size.X,
	  destination,NULL);
    }
    SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
//    SetConsoleTextAttribute(hstdout,(WORD)FG_GREEN);
}
#else
void EscapeFunctionDeleteLine(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect;

    if(ESC_ARG[0]==-1)ESC_ARG[0]=1;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    if(scroll_rect.Top==-1){
	move_rect.Top=buffer_info.dwCursorPosition.Y;
	move_rect.Bottom=buffer_info.srWindow.Bottom;
    }else{
	move_rect.Top=buffer_info.dwCursorPosition.Y;
	move_rect.Bottom=buffer_info.srWindow.Top+scroll_rect.Bottom;
    }
    move_rect.Left=0;
    move_rect.Right=buffer_info.dwSize.X-1;
    destination.X=0;
    destination.Y=move_rect.Top-ESC_ARG[0];
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
    SetConsoleCursorPosition(hstdout,buffer_info.dwCursorPosition);
//    SetConsoleTextAttribute(hstdout,(WORD)FG_GREEN);
}
#endif
void EscapeFunctionScrollUp(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    if(scroll_rect.Top==-1){
//	EscapeFunctionResetScrollArea();
	move_rect.Top=buffer_info.srWindow.Top;
	move_rect.Bottom=buffer_info.srWindow.Bottom;
    }else{
	move_rect.Top=buffer_info.srWindow.Top+scroll_rect.Top;
	move_rect.Bottom=buffer_info.srWindow.Top+scroll_rect.Bottom;
    }
    move_rect.Left=0;
    move_rect.Right=buffer_info.dwSize.X-1;
    destination.X=0;
    destination.Y=move_rect.Top-1;
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
/*
    destination.X=0;
    destination.Y=scroll_rect.Top-1;
    ScrollConsoleScreenBuffer(hstdout,&scroll_rect,
			       &scroll_rect,destination,&char_info);
*/
}

void EscapeFunctionScrollDown(void){
    COORD destination;
    CHAR_INFO char_info;
    SMALL_RECT move_rect;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    if(scroll_rect.Top==-1){
//	EscapeFunctionResetScrollArea();
	move_rect.Top=buffer_info.srWindow.Top;
	move_rect.Bottom=buffer_info.srWindow.Bottom-1;
    }else{
	move_rect.Top=buffer_info.srWindow.Top+scroll_rect.Top;
	move_rect.Bottom=buffer_info.srWindow.Top+scroll_rect.Bottom;
    }
    move_rect.Left=0;
    move_rect.Right=buffer_info.dwSize.X-1;
    destination.X=0;
    destination.Y=move_rect.Top+1;
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
/*
    destination.X=0;
    destination.Y=scroll_rect.Top+1;
    ScrollConsoleScreenBuffer(hstdout,&scroll_rect,
			       &scroll_rect,destination,&char_info);
*/
}

void EscapeFunctionResetScrollArea(void){
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
/*
    scroll_rect.Left=0;
    scroll_rect.Right=buffer_info.dwSize.X-1;
    scroll_rect.Top=buffer_info.srWindow.Top;
    scroll_rect.Bottom=buffer_info.srWindow.Bottom-1;
*/
/*
    scroll_rect.Top=0;
    scroll_rect.Bottom=buffer_info.srWindow.Bottom-buffer_info.srWindow.Top-1;
*/
    scroll_rect.Top=-1;
    scroll_rect.Bottom=-1;
}

void EscapeFunctionSetScrollArea(void){
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    scroll_rect.Left=0;
    scroll_rect.Right=buffer_info.dwSize.X-1;
    if(ESC_ARG[0]!=-1){
//	scroll_rect.Top=buffer_info.srWindow.Top+ESC_ARG[0]-1;
        scroll_rect.Top=ESC_ARG[0]-1;
    }
    if(ESC_ARG[1]!=-1){
//	scroll_rect.Bottom=buffer_info.srWindow.Top+ESC_ARG[1]-1;
        scroll_rect.Bottom=ESC_ARG[1]-1;
    }
/*    printf("%d,%d,%d,%d\n",scroll_rect.Top,scroll_rect.Bottom,
	  scroll_rect.Left,scroll_rect.Right);
*/
}

void EscapeFunctionJISX0208(void){
    ESC_CODING=1;
}

void EscapeFunctionASCII(void){
    ESC_CODING=0;
    ESC_KANJI1=0;
}

void EscapeFunctionHanKata(void){
    ESC_CODING=2;
}

void EscapeFunctionDECPrivateModeSet(void){
    int i;
    COORD dwSize;
    SMALL_RECT srWindow;
    for(i=0;i<ESC_ARGC;i++){
	switch(ESC_ARG[i]){
	case 1:
	    ESC_CURSOR=1;
	    break;
	case 47:
	    GetConsoleScreenBufferInfo(hscreen,&buffer_info);
	    dwSize.X=buffer_info.dwSize.X;
	    dwSize.Y=buffer_info.srWindow.Bottom-buffer_info.srWindow.Top+1;
	    srWindow.Top=0;
	    srWindow.Left=0;
	    srWindow.Bottom=dwSize.Y-1;
	    srWindow.Right=dwSize.X-1;
	    SetConsoleScreenBufferSize(haltscreen,dwSize);
	    SetConsoleWindowInfo(haltscreen,TRUE,&srWindow);
	    SetConsoleScreenBufferSize(haltscreen,dwSize);
	    SetConsoleActiveScreenBuffer(haltscreen);
	    hstdout=haltscreen;
	    break;
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:
	case 38:
	case 40:
	case 41:
	case 44:
	case 45:
	case 46:
	case 1000:
	case 1001:
	    break;
	default:
#ifdef DEBUG
	    fprintf(ESCAPE_LOG,"Unknown ESC [?%dh\n",ESC_ARG[0]);
#endif
	    break;
	}
    }
}

void EscapeFunctionDECPrivateModeReset(void){
    int i;
    for(i=0;i<ESC_ARGC;i++){
	switch(ESC_ARG[0]){
	case 1:
	    ESC_CURSOR=0;
	    break;
	case 47:
	    hstdout=hscreen;
	    SetConsoleActiveScreenBuffer(hstdout);
	    break;
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:
	case 40:
	case 41:
	case 44:
	case 45:
	case 46:
	case 1000:
	case 1001:
	    break;
	default:
#ifdef DEBUG
	    fprintf(ESCAPE_LOG,"Unknown ESC [?%dl\n",ESC_ARG[0]);
#endif
	    break;
	}
    }
}

void EscapeFunctionDeviceStatusReport(void){
    COORD cp;
    char szbuf[16];
    extern SOCKET telnetsocket;
    switch(ESC_ARG[0]){
    case 6:
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	cp.X=buffer_info.dwCursorPosition.X+1;
	cp.Y=buffer_info.dwCursorPosition.Y-buffer_info.srWindow.Top+1;
	sprintf(szbuf,"\x1b[%d;%dR",cp.Y,cp.X);
	send(telnetsocket,szbuf,strlen(szbuf),0);
	break;
    }
}

void EscapeFunctionEnterMode(void){
    if(ESC_ARG[0]==4)ESC_INSERT_MODE=1;
}

void EscapeFunctionEndMode(void){
    if(ESC_ARG[0]==4)ESC_INSERT_MODE=0;
}

void EscapeFunctionSaveCursorPosition(void){
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    CursorPosition=buffer_info.dwCursorPosition;
}

void EscapeFunctionRestoreCursorPosition(void){
    SetConsoleCursorPosition(hstdout,CursorPosition);
}

void EscapeFunctionSetTextParameters(void){
//    printf("SetTextParameters=%d,%s\n",ESC_ARG[0],ESC_TEXT_ARG);
    if(ESC_ARG[0]==2){
	SetConsoleTitle(ESC_TEXT_ARG);
    }
}

void EscapeFunctionDummy(void){}

  /**************************************/
 /*	 EscapeFunctions܂       */
/**************************************/

  /**************************************/
 /*	   䕶֐        */
/**************************************/

void FunctionLineFeed(void){
    COORD coord;
    SMALL_RECT small_rect;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    coord.X=buffer_info.dwCursorPosition.X;
    coord.Y=buffer_info.dwCursorPosition.Y;
    if(coord.Y<buffer_info.srWindow.Bottom-1){
	coord.Y++;
	SetConsoleCursorPosition(hstdout,coord);
    }else if(coord.Y<buffer_info.dwSize.Y-2){
	small_rect=buffer_info.srWindow;
	small_rect.Top++;
	small_rect.Bottom++;
	SetConsoleWindowInfo(hstdout,TRUE,&small_rect);
	coord.Y++;
	SetConsoleCursorPosition(hstdout,coord);
    }else{
	scrollRect2(0,                      // Left
		    0,			    // Top
		    buffer_info.dwSize.X-1, // Right
		    buffer_info.dwSize.Y-2, // Bottom
		    -1);
    }
}

void FunctionBackSpace(void){
    COORD coord;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    coord.Y=buffer_info.dwCursorPosition.Y;
    coord.X=buffer_info.dwCursorPosition.X-1;
    SetConsoleCursorPosition(hstdout,coord);
}

void FunctionHorizontalTab(void){
    COORD coord;
    GetConsoleScreenBufferInfo(hstdout,&buffer_info);
    coord.Y=buffer_info.dwCursorPosition.Y;
    coord.X=(buffer_info.dwCursorPosition.X+8)/8;
    coord.X*=8;
    SetConsoleCursorPosition(hstdout,coord);
}

/*
void writestring(const char * s){
    static DWORD numwritten;
    WriteConsole(hstdout,s,strlen(s),&numwritten,0);
}
*/

void writechar(unsigned char c){
    COORD coord;
    static DWORD numwritten;
    int kanjicode;
    if(ESC_KANJI1!=0){
	if(KANJI_MODE==KM_EUC || ESC_CODING==1){
	    if(ESC_KANJI1==0x8e){
		// EUCp
		outputChar(c);
	    }else{
		// EUCSp or ISO
		kanjicode=((ESC_KANJI1 & 0x7f)<<8)|(c & 0x7f);
		kanjicode=jistosjis(kanjicode);
		outputKanji(kanjicode);
	    }
	}else{
	    // SJIS
	    kanjicode=((ESC_KANJI1<<8)|(unsigned int)c);
	    outputKanji(kanjicode);
	}
	ESC_KANJI1=0;
    }else{
	static int BEL_Count=0;
	if(c!=BEL){
	    BEL_Count=0;
	}
	if(c==BEL){
	    if(BEL_Count!=3){
		BEL_Count++;
		MessageBeep(-1);
	    }
	}else if(c==LF){
	    outputFlash();
#ifdef DEBUG
	    fprintf(ESCAPE_LOG,"LF\n");
#endif
	    FunctionLineFeed();
	}else if(c==HT){
	    outputFlash();
#ifdef DEBUG
	    fprintf(ESCAPE_LOG,"HT\n");
#endif
	    FunctionHorizontalTab();
	}else if(c==BS){
	    outputFlash();
#ifdef DEBUG
	    fprintf(ESCAPE_LOG,"BS\n");
#endif
	    FunctionBackSpace();
	}else if(c==CR){
	    outputFlash();
#ifdef DEBUG
	    fprintf(ESCAPE_LOG,"CR\n");
#endif
	    if(current_line!=-1){
		coord.X=0;
		coord.Y=current_line;
		SetConsoleCursorPosition(hstdout,coord);
	    }else{
		WriteConsole(hstdout,&c,1,&numwritten,0);
	    }
	}else if(c==SI || c==SO){
	}else if(c==0xf2){
	}else{
	    if(ESC_CODING==2){
		outputChar((char)(c | 0x80));
	    }else if(ESC_CODING==1){
		ESC_KANJI1=c;
	    }else{
		if((c & 0x80) && (KANJI_MODE==KM_EUC)){
		    ESC_KANJI1=c;
		}else if(issjis1(c) && (KANJI_MODE==KM_SJIS)){
		    ESC_KANJI1=c;
		}else{
		    outputChar(c);
		}
	    }
	}
    }
}


  /**************************************/
 /*	C^[tF[X       */
/**************************************/


void escinit(void){
#ifdef DEBUG
  ESCAPE_LOG=fopen("EscapeSequence.log","wb");
#endif
}

void escprint(const uchar * s){
    int i;
    i=0;
    while(s[i]!=0)escout(s[i++]);
}

void escout(uchar c){
    //	printf("[%02X]",c);
    if(c==0){
	ESC_FLAG=0;
	return;
    }
    if(ESC_FLAG!=0){
	if(ESC_BYTES==ESC_BUF_SIZE){
	    ESC_FLAG=0;
	}else{
	    ESC_BUF[ESC_BYTES++]=c;
	    LookupEscapeSequence();
	}
    }else if(c!=ESC){
/*
	if(isprint(c) || c==0xd || c==0xa){
	    writechar(c);
	}else{
	    printf("[%02X]",c);
	}
*/
	writechar(c);
    }else{
	ESC_FLAG=1;
	ESC_BYTES=0;
	outputFlash();
    }
	
}

void escSetKanjiMode(int km){
//    if(ESC_KANJI1==0)
    KANJI_MODE=km;
    ESC_KANJI1=0;
    ESC_CODING=0;
}

void escSendSpecialKey(SOCKET s,unsigned char keycode,DWORD controlkeystate){
    unsigned char sendbuf[16];
    unsigned char charcode;
    switch(keycode){
    case VK_LEFT:
    case VK_UP:
    case VK_RIGHT:
    case VK_DOWN:
	if(ESC_CURSOR==0){
	    sendbuf[1]='[';
	}else{
	    sendbuf[1]='O';
	}
	sendbuf[0]=ESC;
	switch(keycode){
	case VK_UP:
	    sendbuf[2]='A';
	    break;
	case VK_DOWN:
	    sendbuf[2]='B';
	    break;
	case VK_RIGHT:
	    sendbuf[2]='C';
	    break;
	case VK_LEFT:
	    sendbuf[2]='D';
	    break;
	}
	send(s,sendbuf,3,0);
	break;
    case VK_END:
    case VK_HOME:
	sendbuf[0]=ESC;
	sendbuf[1]='O';
	switch(keycode){
	case VK_END:
	    sendbuf[2]='F';
	    break;
	case VK_HOME:
	    sendbuf[2]='H';
	    break;
	}
	send(s,sendbuf,3,0);
	break;
    case VK_PRIOR:
    case VK_NEXT:
    case VK_INSERT:
#ifndef DELETE_7F
    case VK_DELETE:
#endif
	sendbuf[0]=ESC;
	sendbuf[1]='[';
	sendbuf[3]='~';
	switch(keycode){
//	case VK_END:
//	    sendbuf[2]='1';
//	    break;
	case VK_INSERT:
	    sendbuf[2]='2';
	    break;
	case VK_DELETE:
	    sendbuf[2]='3';
	    break;
//	case VK_HOME:
//	    sendbuf[2]='4';
//	    break;
	case VK_PRIOR:
	    sendbuf[2]='5';
	    break;
	case VK_NEXT:
	    sendbuf[2]='6';
	    break;
	}
	send(s,sendbuf,4,0);
	break;
#ifdef DELETE_7F
    case VK_DELETE:
	sendbuf[0]=0x7f;
	send(s,sendbuf,1,0);
	break;
#endif
    case VK_F1:
    case VK_F2:
    case VK_F3:
    case VK_F4:
    case VK_F5:
    case VK_F6:
    case VK_F7:
    case VK_F8:
    case VK_F9:
    case VK_F10:
    case VK_F11:
    case VK_F12:
	sendbuf[0]=ESC;
	sendbuf[1]='[';
	sendbuf[4]='~';
	switch(keycode){
	case VK_F1: sendbuf[2]='1'; sendbuf[3]='1'; break;
	case VK_F2: sendbuf[2]='1'; sendbuf[3]='2'; break;
	case VK_F3: sendbuf[2]='1'; sendbuf[3]='3'; break;
	case VK_F4: sendbuf[2]='1'; sendbuf[3]='4'; break;
	case VK_F5: sendbuf[2]='1'; sendbuf[3]='5'; break;
	case VK_F6: sendbuf[2]='1'; sendbuf[3]='7'; break;
	case VK_F7: sendbuf[2]='1'; sendbuf[3]='8'; break;
	case VK_F8: sendbuf[2]='1'; sendbuf[3]='9'; break;
	case VK_F9: sendbuf[2]='2'; sendbuf[3]='0'; break;
	case VK_F10:sendbuf[2]='2'; sendbuf[3]='1'; break;
	case VK_F11:sendbuf[2]='2'; sendbuf[3]='3'; break;
	case VK_F12:sendbuf[2]='2'; sendbuf[3]='4'; break;
	}
	send(s,sendbuf,5,0);
	break;
/*
    case VK_END:
    case VK_HOME:
	sendbuf[0]=ESC;
	sendbuf[1]='O';
	sendbuf[2]=(keycode==VK_HOME)?0:'e';
	send(s,sendbuf,3,0);
	break;
*/
    default:
	if(keycode=='0' && (controlkeystate & SHIFT_PRESSED)){
	    sendbuf[0]='~';
	    send(s,sendbuf,1,0);
	    break;
	}
	   
	charcode=0;
	switch(keycode){
	// L[{[hhCoˑƎvAAA
	case 0xc0: charcode='@'; break;
	case 0xdb: charcode='['; break;
	case 0xdc: charcode='\\'; break;
	case 0xdd: charcode=']'; break;
	case 0xde: charcode='^'; break;
	case 0xe2: charcode='\\'; break;
	}
	if((0x41 <= keycode) && (keycode <= 0x5a))
	  charcode=keycode;
	if(charcode!=0){
	    if((controlkeystate & EITHER_ALT_PRESSED)!=0){
		sendbuf[0]=ESC;
		sendbuf[1]=charcode-0x40;
		send(s,sendbuf,2,0);
	    }else{
		sendbuf[0]=charcode-0x40;
		send(s,sendbuf,1,0);
	    }
	}
/*
	switch(keycode){
	// L[{[hhCoˑƎvAAA
	case 0xc0: keycode='@'; break;
	case 0xdb: keycode='['; break;
	case 0xdc: keycode='\\'; break;
	case 0xdd: keycode=']'; break;
	case 0xde: keycode='^'; break;
	case 0xe2: keycode='\\'; break;
	}
	if((keycode>=0x40) && (keycode<=0x7f)){
	    if((controlkeystate & EITHER_ALT_PRESSED)!=0){
		sendbuf[0]=ESC;
		sendbuf[1]=keycode-0x40;
		send(s,sendbuf,2,0);
	    }else{
		sendbuf[0]=keycode-0x40;
		send(s,sendbuf,1,0);
	    }
	    break;
	}
*/
	break;
    }
}

void escOutputFlush(void){
    FlashCursor = 1;
    outputFlash();
    FlashCursor = 0;
}


  /**************************************/
 /*	        ֐               */
/**************************************/

void LookupEscapeSequence(void){
    int i,j,esc_index,match,fullmatch,anymatch;
/*
    printf("\n{");
    for(i=0;i<ESC_BYTES;i++){
	printf("[%02X]",ESC_BUF[i]);
    }
    printf("}\n");
*/	
    anymatch=0;
    for(esc_index=0;esc_index<ESC_TYPES;esc_index++){
	i=j=0;
	match=fullmatch=0;


	while(1){
	    if(ESC_TABLE[esc_index][i]==1){
		if(isdigit(ESC_BUF[j])){
		    j++;
		    match=1;
		}else{
		    i++;
		    match=0;
		}
	    }else if(ESC_TABLE[esc_index][i]==2){
		if(isprint(ESC_BUF[j])){
		    j++;
		    match=1;
		}else{
		    i++;
		    match=0;
		}
	    }else if(ESC_TABLE[esc_index][i]==ESC_BUF[j]){
	        i++;
		j++;
		match=1;
		if(ESC_TABLE[esc_index][i]==0){
		    fullmatch=1;
		    break;
		}
	    }else{
		match=0;
		break;
	    }
	    if(j==ESC_BYTES)break;
	}


	if(fullmatch!=0)break;		
	anymatch|=match;
    }
/*
    SetConsoleTextAttribute(hstdout,FG_IBLUE);
    ESC_BUF[ESC_BYTES]=0;
    printf("[ESC]%s anymatch=%d\n",ESC_BUF,anymatch);		
    SetConsoleTextAttribute(hstdout,FG_IWHITE);
*/
    if(fullmatch!=0){		// SɃ}b`ꍇ
#ifdef DEBUG
	fprintf(ESCAPE_LOG,"[ESC]");
	for(i=0;i<ESC_BYTES;i++){
	    if(isprint(ESC_BUF[i]))
	      fprintf(ESCAPE_LOG,"%c",ESC_BUF[i]);
	    else
	      fprintf(ESCAPE_LOG,"[%02X]",ESC_BUF[i]);
	}
	fprintf(ESCAPE_LOG,"\n");
#endif      
	//		printf("[Match! No=%d]",esc_index);
	InterpretEscapeSequence(esc_index);
	ESC_FLAG=0;
    }else if(anymatch==0){	// ǂɂ}b`Ȃꍇ
	ESC_FLAG=0;
	ESC_BUF[ESC_BYTES]=0;
#ifdef DEBUG
	//	SetConsoleTextAttribute(hstdout,FG_IRED);
//	printf("[ESC]%s",ESC_BUF);		
	fprintf(ESCAPE_LOG,"[ESC]");
	for(i=0;i<ESC_BYTES;i++){
	    if(isprint(ESC_BUF[i]))
	      fprintf(ESCAPE_LOG,"%c",ESC_BUF[i]);
	    else
	      fprintf(ESCAPE_LOG,"[%02X]",ESC_BUF[i]);
	      
	}
	fprintf(ESCAPE_LOG,"\n");
	//	SetConsoleTextAttribute(hstdout,FG_IWHITE);
#endif
    }
}

void InterpretEscapeSequence(int esc_index){
    int arg_index,i,j,a;
    arg_index=0;
    i=j=0;
    while(ESC_TABLE[esc_index][i]!=0){
	if(ESC_TABLE[esc_index][i]==1){
	    if(isdigit(ESC_BUF[j])){
		a=0;
		do{
		    a*=10;
		    a+=ESC_BUF[j]-'0';
		    j++;
		}while(isdigit(ESC_BUF[j]));
	    }else{
		a=-1;
	    }
	    ESC_ARG[arg_index++]=a;
	    if(arg_index==ESC_NUMARG_TABLE[esc_index])break;
	    i++;
	}else if(ESC_TABLE[esc_index][i]==2){
	    a=0;
	    if(isprint(ESC_BUF[j])){
		do{
		    if(a<ESC_TEXT_ARG_SIZE-1)
		      ESC_TEXT_ARG[a]=ESC_BUF[j];
		    a++;
		    j++;
		}while(isprint(ESC_BUF[j]));
	    }
	    ESC_TEXT_ARG[a]=0;
//	    printf("%s\n",ESC_TEXT_ARG);
	    arg_index++;
	    if(arg_index==ESC_NUMARG_TABLE[esc_index])break;
	    i++;
	}else{
	    i++;
	    j++;
	}
    }
    for(arg_index=ESC_NUMARG_TABLE[esc_index];arg_index<3;arg_index++){
	ESC_ARG[arg_index]=-1;
    }
    ESC_ARGC=ESC_NUMARG_TABLE[esc_index];
    if(ESC_FUNC_TABLE[esc_index]!=NULL){
	ESC_FUNC_TABLE[esc_index]();
    }else{
	SetConsoleTextAttribute(hstdout,FG_IBLUE);
	printf("[ESC %d:",esc_index);		
	for(i=0;i<ESC_NUMARG_TABLE[esc_index];i++){
	    printf("%d",ESC_ARG[i]);
	    if(i!=ESC_NUMARG_TABLE[esc_index]-1){
		printf(",");
	    }
	}
	ESC_BUF[ESC_BYTES]=0;
	printf("(%s)]",ESC_BUF);
	SetConsoleTextAttribute(hstdout,FG_IWHITE);
    }
}




void outputChar(char c){
    OUTPUT_BUF[OUTPUT_BYTES++]=c;
    if(OUTPUT_BYTES==OUTPUT_BUF_SIZE){
        outputFlash();
    }
}

void outputKanji(int i){
    if(OUTPUT_BYTES+2>=OUTPUT_BUF_SIZE)outputFlash();
    OUTPUT_BUF[OUTPUT_BYTES++]=(i>>8);
    OUTPUT_BUF[OUTPUT_BYTES++]=(i & 0xff);
}

void outputFlash(void){
    DWORD numwritten;
    COORD coord; //,destination;
/*
    SMALL_RECT move_rect;
    CHAR_INFO char_info;
*/
    if(OUTPUT_BYTES!=0){
#ifdef DEBUG
      OUTPUT_BUF[OUTPUT_BYTES]=0;
      fprintf(ESCAPE_LOG,"output=%s\n",OUTPUT_BUF);
#endif         
	if(ESC_INSERT_MODE!=0){	// InsertMode
	    ESC_ARG[0]=OUTPUT_BYTES;
	    EscapeFunctionInsertCharacter();
	}
	WriteConsole(hstdout,OUTPUT_BUF,OUTPUT_BYTES,&numwritten,0);
	OUTPUT_BYTES=0;
	GetConsoleScreenBufferInfo(hstdout,&buffer_info);
	coord.Y=buffer_info.dwCursorPosition.Y;
	coord.X=buffer_info.dwCursorPosition.X;
	if(buffer_info.dwCursorPosition.Y==buffer_info.srWindow.Bottom){
	    //ŉsȂXN[B
/*	    move_rect=buffer_info.srWindow;
	    destination.Y=move_rect.Top-1;
	    destination.X=0;
	    char_info.Char.AsciiChar=' ';
	    char_info.Attributes=0;
	    ScrollConsoleScreenBuffer(hstdout,&move_rect,
			       &move_rect,destination,&char_info);
*/
	    scrollRect(&buffer_info.srWindow,-1);
	    coord.Y=buffer_info.dwCursorPosition.Y-1;
	    coord.X=buffer_info.dwCursorPosition.X;
	    SetConsoleCursorPosition(hstdout,coord);
	}else{
	    if(FlashCursor)SetConsoleCursorPosition(hstdout,coord);
	}
	if(coord.X==0)current_line=coord.Y-1;
	else current_line=-1;
    }
}



void scrollRect(SMALL_RECT *move_rect,int lines){
    COORD destination;
    CHAR_INFO char_info;
    destination.X=move_rect->Left;
    destination.Y=move_rect->Top+lines;
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,move_rect,
        move_rect,destination,&char_info);
}

void scrollRect2(int left,int top,int right,int bottom,int lines){
    SMALL_RECT move_rect;
    COORD destination;
    CHAR_INFO char_info;
    move_rect.Left=left;
    move_rect.Top=top;
    move_rect.Right=right;
    move_rect.Bottom=bottom;
    destination.X=left;
    destination.Y=top+lines;
    char_info.Char.AsciiChar=' ';
    char_info.Attributes=0;
    ScrollConsoleScreenBuffer(hstdout,&move_rect,
        &move_rect,destination,&char_info);
}


